#include "mpi.h"
#include <taag.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

/*
extern double drand48();
*/

/* Pipeline version of the algorithm... */
/* we really need the velocities as well... */
typedef struct {
    double x, y, z;
    double mass;
    } Particle;
/* We use leapfrog for the time integration ... */
typedef struct {
    double xold, yold, zold;
    double fx, fy, fz;
    } ParticleV;

void InitParticles( Particle[], ParticleV [], unsigned int, unsigned int, double *, double *, double *);
double ComputeForces( Particle [], ParticleV [], Particle [], unsigned int );
double ComputeNewPos( Particle [], ParticleV [], unsigned int, double, MPI_Comm );

#define XML_DEFAULT "nbodyOnThreeMachine.xml"
#define MAX_PARTICLES_ON_CPU 5 //20 1000
#define DEFAULT_TIME_STEPS 10
#define NR_OF_GALAXIES 3
#define COLLISION_FACTOR 1.4

int main( int argc, char *argv[] )
{
	Particle tmp, centerData[NR_OF_GALAXIES];
	Particle allPoints[18*MAX_PARTICLES_ON_CPU];
	int localRoots[NR_OF_GALAXIES];
	unsigned int collisions[NR_OF_GALAXIES];

    Particle  particles[MAX_PARTICLES_ON_CPU];   /* Particles on ALL nodes */
    ParticleV pv[MAX_PARTICLES_ON_CPU];          /* Particle velocity */
    Particle  sendbuf[MAX_PARTICLES_ON_CPU],     /* buffers */
	recvbuf[MAX_PARTICLES_ON_CPU];
    MPI_Request request[2*NR_OF_GALAXIES], req[2];
    int rank, localRank, size, npart, grpRank, grpSize;
    unsigned int
                 totpart[NR_OF_GALAXIES],   /* total number of particles in each galaxy */
                 indexOfGalaxies[NR_OF_GALAXIES],
                 cnt;                       /* number of times in loop */
    MPI_Datatype particletype;
    int         i, j, pipe, left, right, periodic;
    double      sim_t;                      /* Simulation time */
    double      time;                       /* Computation time */
    MPI_Comm    mpiComm, commring;
    MPI_Group   mpiGrp;
    MPI_Status  statuses[2*NR_OF_GALAXIES], status;

    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    MPI_Comm_size( MPI_COMM_WORLD, &size );

    MPI_Type_contiguous( 4, MPI_DOUBLE, &particletype );
    MPI_Type_commit( &particletype );

    if (argc > 1) {
    	TAAG_Init(argv[1]);
    }
    else {
    	TAAG_Init(XML_DEFAULT);
    }

    if (argc > 2) {
    	cnt = atoi(argv[2]);
    }
    else {
        cnt = DEFAULT_TIME_STEPS;
    }

    TAAG_Group_rank(rank, &grpRank);
    TAAG_Group_MPIGroup(0, NULL, 1, &grpRank, &mpiGrp);
    //MPI_Comm_create is a collective operation
    MPI_Comm_create(MPI_COMM_WORLD, mpiGrp, &mpiComm);
    MPI_Group_rank(mpiGrp, &localRank);
    MPI_Group_size(mpiGrp, &grpSize);

    totpart[grpRank] = MAX_PARTICLES_ON_CPU * size/6*(grpRank+1);
	npart = totpart[grpRank]/grpSize;
	if (totpart[grpRank] % grpSize != 0) {
		npart++;
		totpart[grpRank] = npart * grpSize;
	}

	if (npart > MAX_PARTICLES_ON_CPU) {
		fprintf(stderr,"Error: Too much particles.\n");
		MPI_Finalize();
		return 1;
	}

    //distibutes the global rank of the local roots among the local roots
    if (localRank == 0) {
    	localRoots[grpRank] = rank;
    	if (rank != 0) {
			 //distribute local manager ranks
    		 MPI_Send(&localRoots[grpRank], 1, MPI_INT, 0, grpRank,MPI_COMM_WORLD);
    		 MPI_Recv(localRoots, NR_OF_GALAXIES, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    		 //distribute total particle numbers
    		 MPI_Send(&totpart[grpRank], 1, MPI_UNSIGNED, 0, grpRank,MPI_COMM_WORLD);
    		 MPI_Recv(totpart, NR_OF_GALAXIES, MPI_UNSIGNED, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

    	}
    	else {
    		//collect local manager ranks
    		for (i = 0; i < NR_OF_GALAXIES; i++) {
    			if (i != grpRank) {
    				MPI_Recv(&localRoots[i], 1, MPI_INT, MPI_ANY_SOURCE, i, MPI_COMM_WORLD, &status);
    			}
    		}
    		for (i = 0; i < NR_OF_GALAXIES; i++) {
    			if (i != grpRank) {
    				MPI_Send(localRoots, NR_OF_GALAXIES, MPI_INT, localRoots[i], i ,MPI_COMM_WORLD);
    			}
    		}

    		//collect total particle numbers
    		for (i = 0; i < NR_OF_GALAXIES; i++) {
				if (i != grpRank) {
					MPI_Recv(&totpart[i], 1, MPI_UNSIGNED, MPI_ANY_SOURCE, i, MPI_COMM_WORLD, &status);
				}
			}
			for (i = 0; i < NR_OF_GALAXIES; i++) {
				if (i != grpRank) {
					MPI_Send(totpart, NR_OF_GALAXIES, MPI_UNSIGNED, localRoots[i], i ,MPI_COMM_WORLD);
				}
			}
    	}

    	indexOfGalaxies[0] = 0;
    	for (i = 1; i<NR_OF_GALAXIES; i++) {
    		indexOfGalaxies[i] = indexOfGalaxies[i-1]+totpart[i-1];
    	}
    }

    /* Get the best ring in the topology */
    periodic = 1;
    MPI_Cart_create( mpiComm, 1, &grpSize, &periodic, 1, &commring );
    MPI_Cart_shift( commring, 0, 1, &left, &right );

/* Generate the initial values */
    memset (collisions, 0 , NR_OF_GALAXIES*sizeof(unsigned int));
    //memset (particles, 0 , npart*sizeof(Particle));
    InitParticles( particles, pv, npart, grpRank, &centerData[grpRank].mass, &centerData[grpRank].x, &centerData[grpRank].y);
    tmp.mass = centerData[grpRank].mass;
    tmp.x = centerData[grpRank].x;
    tmp.y = centerData[grpRank].y;
    MPI_Reduce (&tmp.mass, &centerData[grpRank].mass, 1, MPI_DOUBLE, MPI_SUM, 0, commring);
    MPI_Reduce (&tmp.x, &centerData[grpRank].x, 1, MPI_DOUBLE, MPI_SUM, 0, commring);
    MPI_Reduce (&tmp.y, &centerData[grpRank].y, 1, MPI_DOUBLE, MPI_SUM, 0, commring);
    if (localRank == 0) {
    	centerData[grpRank].x = centerData[grpRank].x/centerData[grpRank].mass;
    	centerData[grpRank].y = centerData[grpRank].y/centerData[grpRank].mass;
    }

    time = MPI_Wtime();
    sim_t = 0.0;

    unsigned int cnt_tmp = cnt;
    while (cnt_tmp--) {
    	//distribute the center of mass data among the local roots
    	if (localRank == 0) {
    	    j = 0;
    	    for (i = 0; i < NR_OF_GALAXIES; i++) {
    	    	if (grpRank != i) {
    	    		if (!collisions[i]) {
    	    			MPI_Irecv( &centerData[i], 1, particletype, localRoots[i], MPI_ANY_TAG , MPI_COMM_WORLD, &request[j++] );
						MPI_Isend( &centerData[grpRank], 1, particletype, localRoots[i], grpRank, MPI_COMM_WORLD, &request[j++] );
    	    		} //if
    	    		else {
    	    			//exchange all points
    	    			MPI_Irecv( &allPoints[indexOfGalaxies[i]], totpart[i], particletype, localRoots[i], MPI_ANY_TAG , MPI_COMM_WORLD, &request[j++] );
    	    			MPI_Isend( &allPoints[indexOfGalaxies[grpRank]], totpart[i], particletype, localRoots[i], grpRank, MPI_COMM_WORLD, &request[j++] );
    	    		} //else
    	    	}
    	    }
    	}

    	tmp.mass = 0.0;
    	tmp.x = 0.0;
    	tmp.y = 0.0;
		double max_f, max_f_seg;
		memcpy( sendbuf, particles, npart * sizeof(Particle) );
		max_f = 0.0;

		for (pipe=0; pipe<grpSize; pipe++) {
		     if (pipe != grpSize-1) {
		         MPI_Isend( sendbuf, npart, particletype, right, pipe, commring, &req[0] );
		         MPI_Irecv( recvbuf, npart, particletype, left,  pipe, commring, &req[1] );
		     }
		     // Compute the new center of Mass
		     if (localRank == 0) {
		    	 for (i=0; i<npart; i++) {
		    		 tmp.mass += sendbuf[i].mass;
		    		 tmp.x += sendbuf[i].x;
		    		 tmp.y += sendbuf[i].y;
		    		 allPoints[indexOfGalaxies[grpRank] + pipe* npart +i].mass = sendbuf[i].mass;
		    		 allPoints[indexOfGalaxies[grpRank] + pipe* npart +i].x = sendbuf[i].x;
		    		 allPoints[indexOfGalaxies[grpRank] + pipe* npart +i].y = sendbuf[i].y;
		    	 }
		     }

		     /* Compute forces (2D only) */
		     max_f_seg = ComputeForces( particles, pv, sendbuf, npart );
		     if (max_f_seg > max_f) max_f = max_f_seg;
		     /* Push pipe */
		     if (pipe != grpSize-1) {
		    	 MPI_Waitall( 2, req, statuses );
		    	 memcpy( sendbuf, recvbuf, npart * sizeof(Particle) );
		     }

		}

		//Take into account the center of mass of other galaxies inside each galaxies
		MPI_Waitall( j, request, statuses );
		MPI_Bcast (centerData, NR_OF_GALAXIES, particletype, 0, commring);
		for (i=0; i<NR_OF_GALAXIES; i++) {
			//if (rank == 0) printf("[%f, %f]", centerData[i].x, centerData[i].y); //Print out central coordinates of galaxies
			if (i != grpRank) {
				max_f_seg = ComputeForces( particles, pv, &centerData[i], 1 );
				if (max_f_seg > max_f) max_f = max_f_seg;
			}
		}
		//if (rank == 0) printf("\n"); //The End of Print out central coordinates of galaxies

		/* Once we have the forces, we compute the changes in position */
		sim_t += ComputeNewPos( particles, pv, npart, max_f, mpiComm );

		//distribute the new center of mass data of the galaxies among the local roots
		if (localRank == 0) {
			for (i = 0; i < NR_OF_GALAXIES; i++) {
				if (grpRank != i && !collisions[i]) {
					double dx = (centerData[grpRank].x-centerData[i].x) * (centerData[grpRank].x-centerData[i].x);
					double dy = (centerData[grpRank].y-centerData[i].y) * (centerData[grpRank].y-centerData[i].y);
					//if (rank == 0) printf("The Distance between %d and %d is %f (at step %d).\n",grpRank, i, sqrt(dx+dy), cnt-cnt_tmp);
					if (sqrt(dx+dy) <= COLLISION_FACTOR) {
						printf("Collisions between galaxies %d and %d at the step %d.\n", grpRank, i, cnt-cnt_tmp);
					    collisions[i]++;
					} //if
				}
			}
			centerData[grpRank].mass = tmp.mass;
			centerData[grpRank].x = tmp.x/tmp.mass;
			centerData[grpRank].y = tmp.y/tmp.mass;
		}

		/* We could do graphics here (move particles on the display) */
    } //while
    time = MPI_Wtime() - time;
    if (rank == 0) {
		printf( "Computed %d particles after %d steps in %f seconds\n", MAX_PARTICLES_ON_CPU * size, cnt, time );
    } //if

    MPI_Comm_free(&mpiComm);
    MPI_Comm_free(&commring);
    MPI_Group_free(&mpiGrp);
    TAAG_Free();
    MPI_Finalize();
    return 0;
}

void InitParticles( Particle particles[], ParticleV pv[], unsigned int npart, unsigned int grp, double *sumMass, double *sumCenterX, double *sumCenterY )
{
    unsigned int i;
    *sumMass = 0.0;
    *sumCenterX = 0.0;
    *sumCenterY = 0.0;
    for (i=0; i<npart; i++) {
		particles[i].x	  = drand48()+grp*grp*grp;
		particles[i].y	  = drand48()+grp*grp*grp-2*grp;
		particles[i].z	  = drand48()+grp*grp*grp;
		particles[i].mass = 1.0;
		pv[i].xold	  = particles[i].x;
		pv[i].yold	  = particles[i].y;
		pv[i].zold	  = particles[i].z;
		pv[i].fx	  = 0;
		pv[i].fy	  = 0;
		pv[i].fz	  = 0;

		*sumMass += particles[i].mass;
		*sumCenterX += particles[i].mass * particles[i].x;
		*sumCenterY += particles[i].mass * particles[i].y;
    }
}

double ComputeForces( Particle particles[], ParticleV pv[], Particle sendbuf[], unsigned int npart )
{
  double max_f, rmin;
  int i, j;

  max_f = 0.0;
  for (i=0; i<npart; i++) {
    double xi, yi, rx, ry, mj, r, fx, fy;
    rmin = 100.0;
    xi   = particles[i].x;
    yi   = particles[i].y;
    fx   = 0.0;
    fy   = 0.0;
    for (j=0; j<npart; j++) {
      rx = xi - sendbuf[j].x;
      ry = yi - sendbuf[j].y;
      mj = sendbuf[j].mass;
      r  = rx * rx + ry * ry;
      /* ignore overlap and same particle */
      if (r == 0.0) continue;
      if (r < rmin) rmin = r;
      /* compute forces */
      r  = r * sqrt(r);
      fx -= mj * rx / r;
      fy -= mj * ry / r;
    }
    pv[i].fx += fx;
    pv[i].fy += fy;
    /* Compute a rough estimate of (1/m)|df / dx| */
    fx = sqrt(fx*fx + fy*fy)/rmin;
    if (fx > max_f) max_f = fx;
  }
  return max_f;
}

double ComputeNewPos( Particle particles[], ParticleV pv[], unsigned int npart,
				double max_f, MPI_Comm comm )
{
  int i;
  double      a0, a1, a2;
  static      double dt_old = 0.001, dt = 0.001;
  double      dt_est, dt_new;

  /* integration is a0 * x^+ + a1 * x + a2 * x^- = f / m */
  a0	 = 2.0 / (dt * (dt + dt_old));
  a2	 = 2.0 / (dt_old * (dt + dt_old));
  a1	 = -(a0 + a2);      /* also -2/(dt*dt_old) */

  for (i=0; i<npart; i++) {
    double xi, yi;
    /* Very, very simple leapfrog time integration.  We use a variable
       step version to simplify time-step control. */
    xi	           = particles[i].x;
    yi	           = particles[i].y;
    particles[i].x = (pv[i].fx - a1 * xi - a2 * pv[i].xold) / a0;
    particles[i].y = (pv[i].fy - a1 * yi - a2 * pv[i].yold) / a0;
    pv[i].xold     = xi;
    pv[i].yold     = yi;
    pv[i].fx       = 0;
    pv[i].fy       = 0;
  }

  /* Recompute a time step. Stability criteria is roughly
     2/sqrt(1/m |df/dx|) >= dt.  We leave a little room */
  dt_est = 1.0/sqrt(max_f);
  /* Set a minimum: */
  if (dt_est < 1.0e-6) dt_est = 1.0e-6;
  MPI_Allreduce( &dt_est, &dt_new, 1, MPI_DOUBLE, MPI_MIN, comm );
  /* Modify time step */
  if (dt_new < dt) {
    dt_old = dt;
    dt     = dt_new;
  }
  else if (dt_new > 4.0 * dt) {
    dt_old = dt;
    dt    *= 2.0;
  }

  return dt_old;
}
