/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <taagimpl.h>

/***********************************************************************
 * TAAG_Group_number returns the number of groups.
 *
 ***********************************************************************/
int TAAG_Group_number(int *nr) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	*nr = _taag_getNumberOfGroups();
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Graph requires the maximum size of the square matrix given in
 * the second argument as input returns the distance matrix of
 * the graph of the groups.
 ***********************************************************************/
int TAAG_Group_graph(int maxGroups,  int *distanceMatrix) {
	int nrGroups, copiedAmount;
	int *distance;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	distance = _TAAG_GET_DISTANCES(0);

	if (maxGroups <= nrGroups) {
		copiedAmount = maxGroups;
	} //if
	else {
		copiedAmount = nrGroups;
	} //else

	for (int i = 0; i < copiedAmount; i++) {
		for (int j = 0; j < copiedAmount; j++) {
			*(distanceMatrix + i*nrGroups + j) = *(distance + i*nrGroups +j);
		} //for
	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Group_rank requires a process rank as input and returns the rank
 * of the group which belongs to this process.
 ***********************************************************************/
int TAAG_Group_rank(int rank, int *grpRank) {
	int maxProcs, nrGroups;
	int *nrProcs;
	int *processes;
	int found = TAAG_FALSE;

	*grpRank = -1;
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
    if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

    nrGroups = _taag_getNumberOfGroups();

    for (int i = 0; i < nrGroups && !found; i++) {
    	nrProcs = _TAAG_GET_NR_OF_PROCESSES(i);
    	processes = *( _TAAG_GET_LIST_OF_PROCESSES(i) );

    	for (int j = 0; j < *nrProcs && !found; j++) {
    		if (rank == *processes) {
    			*grpRank = i;
    			found = TAAG_TRUE;
    		} //if
    		else {
    			processes++;
    		} //else
    	} //for

    } //for

    if (found) return (TAAG_SUCCESS);
    else return (TAAG_ERR_UNKNOWN);
}

/***********************************************************************
 * TAAG_Group_size requires a group rank as input and returns the
 * number of member processes of the group.
 ***********************************************************************/
int TAAG_Group_size(int grpRank, int *nr) {
	int nrGroups;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank < 0 || grpRank >= nrGroups) return (TAAG_ERR_GRANK);

	*nr = *( _TAAG_GET_NR_OF_PROCESSES(grpRank) );

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Group_members requires a group rank and the maximum size of the
 * vector given in the third argument as input and returns
 * the ranks of all member processes of the group.
 ***********************************************************************/
int TAAG_Group_members(int grpRank, int nr, int *members) {
	int *nrProcs;
	int *processes;
	int nrGroups, copiedAmount;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank < 0 || grpRank >= nrGroups) return (TAAG_ERR_GRANK);

	nrProcs = _TAAG_GET_NR_OF_PROCESSES(grpRank);
	processes = *( _TAAG_GET_LIST_OF_PROCESSES(grpRank) );

	if (nr <= *nrProcs) {
		copiedAmount = nr;
	} //if
	else {
		copiedAmount = *nrProcs;
	} //else

	for (int i = 0; i < copiedAmount; i++) {
		*(members+i) = *(processes+i);
	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Group_MPIGroup requires the number and the enumeration of
 * some process ranks; furthermore the number and the enumeration of
 * some group ranks as input and it composes a MPI_Group
 * from the given processes and groups.
 *
 * Comments: This call is useful if some MPI collective operations are
 * going to be used within among the given processes or groups.
 *
 * The order of processes in the MPI_Group object is: given processes
 * ﬁrst (in the given order) then the members of the given groups
 * (in the given order of groups).
 ***********************************************************************/
int TAAG_Group_MPIGroup(int nrProcs, int *ranks, int nrGroups, int *grpRanks, MPI_Group *grp) {
	int maxGroups;
	int maxProcs;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (nrProcs < 0 || nrProcs > maxProcs) return (TAAG_ERR_COUNT);

	maxGroups = _taag_getNumberOfGroups();
	if (nrGroups < 0 || nrGroups > maxGroups) return (TAAG_ERR_GRANK);

	// **** Checking process ranks ****
		for (int i = 0; i < nrProcs; i++) {
			if ( *(ranks+i) < 0 || *(ranks+i) >= maxProcs) return (TAAG_ERR_PRANK);
		} //for

	// **** Checking group ranks ****
	for (int i = 0; i < nrGroups; i++) {
		if ( *(grpRanks+i) < 0 || *(grpRanks+i) >= maxGroups) return (TAAG_ERR_GRANK);
	} //for

	//------------------------------------------------------------------
	MPI_Group orig_group;
	MPI_Group inProcesses;
	MPI_Group inGroups;
	int processes[maxProcs];

	int *sizeOfGroup;
	int *procsOfGroup;

	MPI_Comm_group(MPI_COMM_WORLD, &orig_group);

	//***** MPI_Group_incl works even if nrProcs == 0
	MPI_Group_incl(orig_group, nrProcs, ranks, &inProcesses);

	//***** MPI_Group_incl works even if nrGroups == 0
	int counter = 0;
	for (int i = 0; i < nrGroups; i++) {
		sizeOfGroup = _TAAG_GET_NR_OF_PROCESSES(*(grpRanks+i));
		procsOfGroup = *( _TAAG_GET_LIST_OF_PROCESSES(*(grpRanks+i)) );
		for (int j = 0; j < *sizeOfGroup; j++) {
			processes[counter++] = *(procsOfGroup+j);
		} //for
	} //for
	MPI_Group_incl(orig_group, counter, processes, &inGroups);

	/*****In the union the order of the *****/
	/***** input groups is important *****/
	MPI_Group_union(inProcesses, inGroups, grp);
	MPI_Group_free(&inProcesses);
	MPI_Group_free(&inGroups);

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Groups_degree requires a group rank and the order (distance)
 * of the queried neighbors as input. If the
 * execution plan defines which groups are planned to interact each other
 * then this call returns the number of the n-th order “neighbor” groups
 * of the given one.
 *
 * Comments: If the second argument is equal to:
 * 		0: the group itself (no reason)
 * 		1: 1st order neighbors
 * 		2: 2st order neighbors
 * 		...
 *     -1: groups which are unreachable from the given one
 ***********************************************************************/
int TAAG_Group_degree(int grpRank, int order, int *nr) {
	int nrGroups;
	int *distance;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank < 0 || grpRank >= nrGroups) return (TAAG_ERR_GRANK);

	if (order < -1) return (TAAG_ERR_ARG);

	distance = _TAAG_GET_DISTANCES(grpRank);

	*nr = 0;
	for (int i = 0; i < nrGroups; i++) {
		if (*(distance+i) == order) {
			(*nr)++;
		}
    } //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Groups_neighbours requires a group rank, the order (distance)
 * of the queried neighbors and the maximum size of the vector given in
 * the fourth argument and as input. If the execution plan defines which
 * groups are planned to interact each other
 * (e.g.: in the case of a schemas tree and ring), then this call
 * returns the ranks of the n-th order “neighbor” groups of the given one.
 *
 * comments: if the third argument is equal to -1, then the call retruns
 * the ranks of the unreachable groups (see TAAG_Group_degree).
 ***********************************************************************/
int TAAG_Group_neighbours(int grpRank, int order, int nr, int *list) {
	int *distance;
	int nrGroups, copiedAmount;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank < 0 || grpRank >= nrGroups) return (TAAG_ERR_GRANK);

	if (order < -1) return (TAAG_ERR_ARG);

	distance = _TAAG_GET_DISTANCES(grpRank);

	if (nr <= nrGroups) {
		copiedAmount = nr;
	} //if
	else {
		copiedAmount = nrGroups;
	} //else

	int j = 0;
	for (int i = 0; i < nrGroups && j < copiedAmount; i++) {
		if (*(distance+i) == order) {
			*(list+j) = i;
			j++;
		} //if
	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Group_distance requires ranks of two groups as input and
 * returns the number of edges (the shortest distance) between the two
 * groups in the predefined graph.
 * If there is now a predefined way between the two groups in the graph
 * the the third argument returns the value -1.
 ***********************************************************************/
int TAAG_Group_distance(int grpRank1, int grpRank2, int *nr) {
	int nrGroups;
	int *distance;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank1 < 0 || grpRank1 >= nrGroups) return (TAAG_ERR_GRANK);
	if (grpRank2 < 0 || grpRank2 >= nrGroups) return (TAAG_ERR_GRANK);

	distance = _TAAG_GET_DISTANCES(grpRank1);

	*nr = *(distance + grpRank2);

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Group_way requires ranks of two groups and the maximum size of
 * the vector given in the forth argument as input and
 * returns the list of the ranks of the intermediate groups (along the
 * shortest path) with grpRank2 at the end.
 * If there is now a predefined way between the two groups in the graph
 * the forth one returns a NULL.
 ***********************************************************************/
int TAAG_Group_way(int grpRank1, int grpRank2, int nr, int *list) {
	int nrGroups;
	int  currentDist, copiedAmount;
	int *distance;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank1 < 0 || grpRank1 >= nrGroups) return (TAAG_ERR_GRANK);
	if (grpRank2 < 0 || grpRank2 >= nrGroups) return (TAAG_ERR_GRANK);

	distance = _TAAG_GET_DISTANCES(0);

	currentDist = *(distance + grpRank1 * nrGroups + grpRank2);
	if (currentDist == -1 || currentDist == 0) return (TAAG_SUCCESS);

	if (nr <= currentDist) {
			copiedAmount = nr;
		} //if
		else {
			copiedAmount = currentDist;
		} //else

	for (int i = 0; i < copiedAmount && grpRank1 != grpRank2 ; i++ ) {
		int j = 0;
		while ( 	!(
					*(distance + grpRank1 * nrGroups + j) == 1
						&&
					*(distance + j * nrGroups + grpRank2) < currentDist
					)) {
			j++;
		} //while
		grpRank1 = *(list+i) = j;
		currentDist = *(distance + j * nrGroups + grpRank2);
	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Group_element requires a group rank as input and integer n
 * number (where n is from 0 to less by one than the number of processes
 * in the group) and returns the rank of the n –th process in the given
 * group.
 ***********************************************************************/
int TAAG_Group_element(int grpRank, int index, int *rank) {
	int *nrProcs;
	int *processes;
	int nrGroups;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	nrGroups = _taag_getNumberOfGroups();
	if (grpRank < 0 || grpRank >= nrGroups) return (TAAG_ERR_GRANK);

	nrProcs = _TAAG_GET_NR_OF_PROCESSES(grpRank);
	processes = *( _TAAG_GET_LIST_OF_PROCESSES(grpRank) );

	if (index < 0 || index >= *nrProcs) return (TAAG_ERR_COUNT);

	*rank = *(processes+index);

	return (TAAG_SUCCESS);
}
