/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <taagimpl.h>

/***********************************************************************
 * TAAG_Topology_given indicates whether given execution plan (ﬁle)
 * describes a network topology. It returns a ﬂag as either logical
 * true (1) or false(0).
 ***********************************************************************/
int TAAG_Topology_given(int *flag) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	*flag = _hw_initialized;
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_CPU_number returns the number of the allocated CPUs
 *
 ***********************************************************************/
int TAAG_CPU_number(int *nr) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	*nr = _taag_nrOfCPUs();
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Host_number returns the number of the allocated hosts
 *
 ***********************************************************************/
int TAAG_Host_number(int *nr) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	*nr = _taag_nrOfHost();
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Host_properites requires a process rank as input and returns the
 * number of CPUs and processes on the host where the given
 * process is located.
 *
 * Comment: NULL pointer can be applied for the second and third
 * argument.
 ***********************************************************************/
int TAAG_Host_properties(int rank, int *nrCPUs, int *nrProcs) {
	int host;
    int maxProcs;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
    if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

    host = _taag_getHost(rank);
    if (host < 0) return (TAAG_ERR_UNKNOWN);

	if (nrCPUs != NULL) {
		*nrCPUs = *( _TAAG_GET_NR_CPUS(host) );
	}
	if (nrProcs != NULL) {
		*nrProcs = *( _TAAG_GET_NR_PROCS_ON_HOST(host) );
	}

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Host_processes requires a process rank and the maximum number of
 * the vector given in the third argument as input and it returns
 * the ranks of all the processes residing on the same host.
 ***********************************************************************/
int TAAG_Host_processes(int rank, int nr, int *ranks) {
	int *nrProcs;
	int *processes;
	int copiedAmount;
	int host;
	int maxProcs;
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
    if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

    host = _taag_getHost(rank);
    if (host < 0) return (TAAG_ERR_UNKNOWN);

    nrProcs = _TAAG_GET_NR_PROCS_ON_HOST(host);
    processes = *( _TAAG_GET_LIST_OF_PROCS_ON_HOSTS(host) );

    if (nr <= *nrProcs) {
   		copiedAmount = nr;
   	} //if
   	else {
   		copiedAmount = *nrProcs;
   	} //else

   	for (int i = 0; i < copiedAmount; i++) {
   		*(ranks+i) = *(processes+i);
    } //for

    return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Host_latency requires the ranks of two processes as input and
 * returns the average latency value between two hosts on which
 * the given processes reside.
 *
 * Comment: If the given two processes are located on the same host,
 * then the call returns with 0.
 ***********************************************************************/
int TAAG_Host_latency(int rank1, int rank2, double *latency) {
	int host1, host2;
	int maxProcs;
	double *latencies;
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
    if (rank1 < 0 || rank1 >= maxProcs) return (TAAG_ERR_PRANK);
    if (rank2 < 0 || rank2 >= maxProcs) return (TAAG_ERR_PRANK);

    host1 = _taag_getHost(rank1);
    host2 = _taag_getHost(rank2);
    if (host1 < 0 || host2 < 0) return (TAAG_ERR_UNKNOWN);

    if (host1 == host2) {
    	*latency = 0.0;
    } //if
    else {
    	latencies = _TAAG_GET_LATENCIES(host1);
    	*latency = *(latencies + host2);
    } //else

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Host_address requires a process rank and the maximum size of the
 * vector given in the third argument as input and returns the
 * address of the host where the process resides.
 *
 * Comment: The size of the given char vector should be
 * TAAG_MAX_HOSTNAME_STRING
 ***********************************************************************/
int TAAG_Host_address (int rank, char *address) {
	int host;
	int maxProcs;
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
    if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

    host = _taag_getHost(rank);
    if (host < 0) return (TAAG_ERR_UNKNOWN);

    strcpy(address, *( _TAAG_GET_HOSTNAMES(host) ) );

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_LAN_number returns the number of LAN whose resources are used
 * in the current session
 ***********************************************************************/
int TAAG_LAN_number(int *nr) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	*nr = _taag_nrOfLAN();
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_LAN_properties requires a a process rank as input and returns
 * the number of hosts, the number of CPUs and the number of processes in
 * the LAN where the given process is located.
 *
 * Comment: NULL pointer can be applied  for the last three arguments.
 ***********************************************************************/
int TAAG_LAN_properties(int rank, int *nrHosts, int *nrCPUs, int *nrProcs) {
	int maxProcs;
	int host;
	int lan;
	int curHosts, prevHosts;
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
    if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

    host = _taag_getHost(rank);
    if (host < 0) return (TAAG_ERR_UNKNOWN);

    lan = _taag_getLAN(host);
    if (lan < 0) return (TAAG_ERR_UNKNOWN);

    curHosts = *( _TAAG_GET_HOSTS(lan) );
    if (lan == 0) {
    	prevHosts = 0;
    } //if
    else {
    	prevHosts = *( _TAAG_GET_HOSTS(lan -1) );
    } //else

    if (nrHosts != NULL) {
    	*nrHosts = curHosts - prevHosts;
    }


    if (nrCPUs != NULL) {
    	*nrCPUs = 0;
    	for (int i = prevHosts; i < curHosts; i++) {
    		*nrCPUs += *( _TAAG_GET_NR_CPUS(i) );
    	} //for
    } //if

	if (nrProcs != NULL) {
		*nrProcs = 0;
		for (int i = prevHosts; i < curHosts; i++) {
			*nrProcs += *( _TAAG_GET_NR_PROCS_ON_HOST(i) );
		} //for
	} //if
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_LAN_host requires a a process rank and the size of the vector
 * given in the third arguments as input and returns the
 * ranks of hosts residing in the LAN where the given process is located.
 ***********************************************************************/
//int TAAG_LAN_host(int rank, int nr, int *hosts) {
//	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
//	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);
//
//	return (TAAG_SUCCESS);
//}

/***********************************************************************
 * TAAG_LAN_processes requires a process rank and the maximum number of
 * the vector given in the third argument as input and it returns the
 * ranks of those processes which are executed on the same LAN.
 ***********************************************************************/
int TAAG_LAN_processes(int rank, int nr, int *ranks) {
	int *nrProcs;
	int *processes;
	int host;
	int lan;
	int maxProcs;
	int prevHosts, curHosts;
	int finished = TAAG_FALSE;
	int counter;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	host = _taag_getHost(rank);
	if (host < 0) return (TAAG_ERR_UNKNOWN);

	lan = _taag_getLAN(host);
	    if (lan < 0) return (TAAG_ERR_UNKNOWN);

	curHosts = *( _TAAG_GET_HOSTS(lan) );
	if (lan == 0) {
		prevHosts = 0;
	} //if
	else {
		prevHosts = *( _TAAG_GET_HOSTS(lan -1) );
	} //else

	counter = 0;
	for (int i = prevHosts; i < curHosts && !finished; i++) {
		nrProcs = _TAAG_GET_NR_PROCS_ON_HOST(i);
		processes = *( _TAAG_GET_LIST_OF_PROCS_ON_HOSTS(i) );


		for (int j = 0; j < *nrProcs && !finished; j++) {
			if (counter < nr) {
				*(ranks+counter) = *(processes+j);
				counter++;
			} //if
			else {
				finished = TAAG_TRUE;
			} //else
		} //for
	} //for
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Comm_level requires two process ranks as input and returns on
 * which network level they can communicate with each other.
 *     - If commLevel = 0 then the two processes can interact each other
 *       only via WAN,
 *     - but if commLevel = 1 they are located in the same LAN network
 *     - and if commLevel = 2 they nest on the same host.
 ***********************************************************************/
int TAAG_Comm_level(int rank1, int rank2, int *commlevel) {
	int host1, host2;
	int lan1, lan2;
	int maxProcs;
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (_hw_initialized == TAAG_FALSE) return (TAAG_ERR_HWTOPOLOGY);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank1 < 0 || rank1 >= maxProcs) return (TAAG_ERR_PRANK);
	if (rank2 < 0 || rank2 >= maxProcs) return (TAAG_ERR_PRANK);

	host1 = _taag_getHost(rank1);
	host2 = _taag_getHost(rank2);
	if (host1 < 0 || host2 < 0) return (TAAG_ERR_UNKNOWN);

	if (host1 == host2) {
		*commlevel = TAAG_HOST_LEVEL;
	} //if
	else {
		lan1 = _taag_getLAN(host1);
		lan2 = _taag_getLAN(host2);
		if (lan1 < 0 || lan2 < 0) return (TAAG_ERR_UNKNOWN);

		if (lan1 == lan2) {
			*commlevel = TAAG_LAN_LEVEL;
		} //if
		else {
			*commlevel = TAAG_WAN_LEVEL;
		} //else
	} //else
	return (TAAG_SUCCESS);
}
