/*
 * $Id$
 * Author: Karoly Bosa
 *
 * (C) 2008 by Research Institute for Symbolic Computation, Linz, Austria
 * All rights reserved. See COPYRIGHT in top-level directory.
 */

#include <taagimpl.h>

/***********************************************************************
 * TAAG_Tree_isTree indicates whether given execution plan
 * (ﬁle) describes a tree structure. It returns a ﬂag as either logical
 * true (1) or false(0).
 ***********************************************************************/
int TAAG_Tree_isTree(int *flag) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);

	if (SCHEMA_TREE != _taag_getType()) {
		*flag = TAAG_FALSE;
	}
	else {
		*flag = TAAG_TRUE;
	}
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_depth returns the depth (number of the levels) of the tree.
 *
 ***********************************************************************/
int TAAG_Tree_depth(int *depth) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	*depth = _taag_getDepth();
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_level requires a process rank and returns on which level
 * the given process is located on the tree. The level of
 * the root is 0 and the level of the leaves is depth − 1.
 ***********************************************************************/
int TAAG_Tree_level(int rank, int *level) {
	int maxProcs, grpRank;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	TAAG_Group_rank(rank, &grpRank);
	TAAG_Group_distance(TAAG_ROOT, grpRank, level);

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_width requires
 * 		- process rank which specifies root of a subtree,
 *      - a level in the specified subtree (or degree of children) and
 * it returns the width of the subtree on the given level in terms of
 * processes (or 0 if the given level number is not applicable on the
 * given subtree).
 *
 * Comments: the tree structure described by the execution plan may be
 * unbalanced, it can therefore be useful to know e.g.: the number of
 * the available workers on particular subtree).
 ***********************************************************************/
int TAAG_Tree_width(int rank, int level, int *nr) {
	int maxProcs;
	int grpRank, myLevel;
	int nrGroupsOnLevel;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	*nr = 0;
	TAAG_Group_rank(rank, &grpRank);
	TAAG_Group_distance(TAAG_ROOT, grpRank, &myLevel); //or TAAG_Tree_level(rank, &myLevel);
	TAAG_Group_degree(grpRank, level, &nrGroupsOnLevel);

	int neighbours[nrGroupsOnLevel];
	TAAG_Group_neighbours(grpRank, level, nrGroupsOnLevel, neighbours);

	for (int i = 0; i < nrGroupsOnLevel; i++) {
        int levelOfNeighbour;
        int groupSize;
        TAAG_Group_distance(TAAG_ROOT, neighbours[i], &levelOfNeighbour);

        if (myLevel + level != levelOfNeighbour) continue;

        TAAG_Group_size(neighbours[i], &groupSize);

        *nr += groupSize;
	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_children requires
 * 		- process rank which specifies root of a subtree,
 *      - a level in the specified subtree (or degree of children) and
 *      - and the maximum size of the vector given in the forth argument
 * and returns the rank of the successor processes located on the given
 * level in the specified subtree.
 ***********************************************************************/
int TAAG_Tree_children(int rank, int level, int nr, int *procs) {
	int maxProcs;
	int grpRank, myLevel;
	int nrGroupsOnLevel, counter;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	TAAG_Group_rank(rank, &grpRank);
	TAAG_Group_distance(TAAG_ROOT, grpRank, &myLevel); //or TAAG_Tree_level(rank, &myLevel);
	TAAG_Group_degree(grpRank, level, &nrGroupsOnLevel);

	int neighbours[nrGroupsOnLevel];
	TAAG_Group_neighbours(grpRank, level, nrGroupsOnLevel, neighbours);

	counter = 0;
	for (int i = 0; i < nrGroupsOnLevel && counter < nr; i++) {
        int levelOfNeighbour;
        int groupSize;
        TAAG_Group_distance(TAAG_ROOT, neighbours[i], &levelOfNeighbour);

        if (myLevel + level != levelOfNeighbour) continue;

        TAAG_Group_size(neighbours[i], &groupSize);

        int processes[groupSize];
        TAAG_Group_members(neighbours[i], groupSize, processes);

        for (int j = 0; j < groupSize && counter < nr; j++) {
        	*(procs+counter) = processes[j];
        	counter++;
        } //for

	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_parent requires a process rank and returns the rank of the
 * parent (process) of the given process.
 ***********************************************************************/
int TAAG_Tree_parent(int rank, int *parent) {
	int maxProcs;
	int grpRank, myLevel;
	int nrGroupsOnLevel;
	int found = TAAG_FALSE;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	TAAG_Group_rank(rank, &grpRank);
	TAAG_Group_distance(TAAG_ROOT, grpRank, &myLevel); //or TAAG_Tree_level(rank, &myLevel);
	TAAG_Group_degree(grpRank, 1, &nrGroupsOnLevel);

	int neighbours[nrGroupsOnLevel];
	TAAG_Group_neighbours(grpRank, 1, nrGroupsOnLevel, neighbours);

	*parent = TAAG_ROOT;

	for (int i = 0; i < nrGroupsOnLevel && !found; i++) {
        int levelOfNeighbour;
        TAAG_Group_distance(TAAG_ROOT, neighbours[i], &levelOfNeighbour);
        if (myLevel - 1 == levelOfNeighbour) {
        	TAAG_Group_element(neighbours[i], 0, parent);
        	found = TAAG_TRUE;
        } //if

	} //for

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_root returns the rank of the root process.
 *
 * Comment: The rank of the root group (as well as the rank of the root
 * process) is always TAAG_ROOT (0).
 ***********************************************************************/
int TAAG_Tree_root(int *rank) {
	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	TAAG_Group_element(TAAG_ROOT, 0, rank);
	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_isLeaf requires a process rank and indicates whether the
 * given process is located on the level on depth−1 in the tree. The
 * call returns a ﬂag as either logical true (1) or false (0).
 ***********************************************************************/
int TAAG_Tree_isLeaf(int rank, int *flag) {
	int maxProcs;
	int depth, myLevel;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	TAAG_Tree_depth(&depth);
	TAAG_Tree_level(rank, &myLevel);

	if (myLevel + 1 == depth) {
		*flag = TAAG_TRUE;
	}
	else {
		*flag = TAAG_FALSE;
	}

	return (TAAG_SUCCESS);
}

/***********************************************************************
 * TAAG_Tree_isLocalManager requires a process rank and indicates
 * whether the given process is located on the level on depth−2 in the
 * tree. The call returns a ﬂag as either logical true (1) or false (0).
 *
 * Comments: The local managers are specials in the sense that they are
 * always located on the same local physical infrastructure (Host or LAN)
 * as their children.
 ***********************************************************************/
int TAAG_Tree_isLocalManager(int rank, int *flag) {
	int maxProcs;
	int depth, myLevel;

	if (_taag_initialized == TAAG_FALSE) return (TAAG_ERR_INIT);
	if (SCHEMA_TREE != _taag_getType()) return (TAAG_ERR_SCHEMA);

	MPI_Comm_size(MPI_COMM_WORLD, &maxProcs);
	if (rank < 0 || rank >= maxProcs) return (TAAG_ERR_PRANK);

	TAAG_Tree_depth(&depth);
	TAAG_Tree_level(rank, &myLevel);

	if (myLevel + 2 == depth) {
		*flag = TAAG_TRUE;
	}
	else {
		*flag = TAAG_FALSE;
	}

	return (TAAG_SUCCESS);
}
